'''
Created on 18 Jan 2011

@author: Simon Wilshin, swilshin@rvc.ac.uk

These are the base classes for the analysis plugins. When 
a new analysis is required these should be inherited from. 
The procedure for writing a new analysis is:
  1. Inherit from either the UnsampledBaseAnalysis or BaseAnalysis 
     class contingent on if your new analysis is sampled or not.
  2. Overide the default targets to point to the elements of the 
     experiment tree you would like to access by default.
  3. Write a new constructor which resolves these targets and 
     calls the base class constructor
  4. If apropriate write the keys() and has_key() methods so 
     that higher levels of the heirachy can reveal the relevant 
     interfaces.
'''

import conventions as conv
from directoryAware import DirectoryAware as DirectoryAware 
from tree import Leaf
from defaultable import Defaultable
from mDInterpret import getID, loadMetaData

class UnsampledBaseAnalysis(DirectoryAware,Leaf,Defaultable):
    '''
    Base class for all analyses. Is a leaf in the experiment tree, can 
    have default settings from the metadata and knows the directories it 
    should read and write to. Inherit analyses from this if your analysis 
    is unsampled (for example marker set information). Otherwise used the 
    BaseAnalysis class.
    '''
    defaultTargets = None # Resolved using the resolve() method the tree classes.
    def __init__(self, workingDirectory, analysisDirectory, metaData, parent):
        '''
        Construct the analysis object with apropriate knowledge of directories 
        it can operate on, and with suitable defaults and treeId, as well as the 
        parent object.
        @param workingDirectory: The working directory relative to which this objects 
        working directory should be resolved
        @type workingDirectory: string
        @param workingDirectory: The working directory relative to which this objects 
        working directory should be resolved
        @type workingDirectory: string
        @param metaData: Standard metadata dictionary.
        @type metaData: dict
        @param metaData: Parent in the experiment tree.
        @type metaData: ExperimentBranch,
        '''
        if not getID(metaData):
            metaData["iD"] = parent.iD + type(self).defaultIDExt
        DirectoryAware.__init__(self, workingDirectory,analysisDirectory,metaData)
        # Set the targets for this analysis
        if metaData.has_key("targets"):
            self.targets = metaData["targets"]
        else:
            self.targets = type(self).defaultTargets
        if metaData.has_key("iD"):
            Leaf.__init__(self,parent,metaData["iD"])
        else:
            Leaf.__init__(self,parent)
        Defaultable.__init__(self, metaData)

class BaseAnalysis(UnsampledBaseAnalysis):
    '''
    Base class for all analyses which are sampled, that is have an asociated 
    time. Is a leaf in the experiment tree, can have default settings 
    from the metadata and knows the directories it should read and 
    write to. Inherit analyses from this if your analysis 
    is unsampled (for example marker set information). Otherwise used the 
    BaseAnalysis class.
    '''
    def __init__(self, workingDirectory, analysisDirectory, metaData, parent):
        '''
        Constructor, initialise the associated file, sample rate, etc. The 
        dictionary data is where all computed quantities should be stored. 
        These values can be obtained by indexing the object with a suitable 
        key.
        @param workingDirectory: The working directory relative to which this objects 
        working directory should be resolved
        @type workingDirectory: string
        @param workingDirectory: The working directory relative to which this objects 
        working directory should be resolved
        @type workingDirectory: string
        @param metaData: Standard metadata dictionary.
        @type metaData: dict
        @param parent: Parent in the experiment tree.
        @type parent: ExperimentBranch,
        '''
        UnsampledBaseAnalysis.__init__(self, workingDirectory, analysisDirectory, metaData, parent)
        loadMetaData("sampleRate",self, self.metaData,None) # Can be specified later
        loadMetaData("timeCuts",self, self.metaData,None)
        self.totalSamples=None
        self.data=dict()
        
    def __setitem__(self, k, v):
        '''
        Virtual method to allow item assignment.
        @param k: key
        @type k: any immutable,
        @param v: value
        @type k: any
        '''
        self.data[k]=v
        
    def mangleKey(self,k):
        '''
        Virtual method, when overridden should convert the key k to a
        sensible entry in the dictionary self.data. The idea is to make 
        the system robust by having it do something sensible regardless 
        of how the user request the data be given.
        Returns: A mangled version of the key k. 
        @param k: key
        @type k: any immutable,
        '''
        return(k)

    def has_key(self,k):
        '''
        Checks if the key k exists in the data object. The intention is to 
        allow users to query if a key exists interactively so the key is 
        mangled (projected from a list of acceptable key names on to the 
        space of actual key names in data) first. As such this should not 
        be used for the purposes of finding unique keys.
        Returns: If the key k corresponds to data in the object
        @param k: key
        @type k: any immutable,
        '''
        return(self.data.has_key(self.mangleKey(k)))

    def __getitem__(self,k):
        '''
        Returns an element of the dictionary self.data based on the key k 
        which is mangled (projected from a list of acceptable key names on 
        to the space of actual key names in data) first so the user has a 
        variety of sensible ways to get at the data. Will throw an exception 
        if k does not correspond to anything after mangling.
        Returns: The data associated with k. 
        @param k: key
        @type k: any immutable,
        '''
        return(self.data[self.mangleKey(k)])

    def keys(self):
        '''
        Returns a list of unique keys for this data object. This is a list 
        of unique keys not a list of keys including mangling options. No mangling
        (projecting from a list of acceptable key names on to the space of actual
        key names in data) is performed here. 
        '''
        return(self.data.keys())

if __name__=='__main__':
    '''
    No tests here since the classes here are never supposed to be instantiated.
    '''
    print "No tests for baseData, these classes should never be directly instantiated"