'''
Created on 06 Oct 2011

@author: Simon Wilshin, swilshin@rvc.ac.uk

Computes the location of the CoM for a motion capture data
set using the body markers defined in a marker set.
'''

import numpy as np
from numpy import nanstd
#from matplotlib.mlab import prepca
from sklearn import decomposition
from os.path import join as osjoin
from os import listdir as listdir
import pickle

from baseAnalysis import BaseAnalysis
from utils import GPInterpolater
from mDInterpret import getID, loadMetaData
from scipy.stats import linregress

class CoM(BaseAnalysis):
    '''
    Finds the center of mass of the body markers of some motion 
    capture measurement. Default targets are a local MarkerSet 
    object and a local MotionCaptureData object, it gets the 
    list of body markers from the MarkerSet, the positions 
    from the MotionCaptureData and averages using a gaussian 
    process. Considerable tweaking of the parameters of this 
    object are necessary to get the GP to fill properly at 
    the moment. Through the metaData one can change
    @param numPC: Number of principle components for the 
    x co-ordinates of the GP, more provides more information 
    to the GP, but will tend to drag interpolation towards 
    the prior, that is mean zero.
    @type numPC: int
    @param Nmeas: Number of measurements to use for the GP 
    interpolation. More will be slower, but provide better 
    interpolation
    @type Nmeas: int
    @param cached: Should the result of the CoM estimation 
    be cached in the analysis directory so that it can be 
    loaded rather than computed next time (good if the GP 
    interpolation is slow).
    @type cached: bool
    @param ardSupression: Automatic relevance determination 
    supression. ARD tends to ignore spatial co-ordinates in 
    preference to time co-ordinates, with this you can supress 
    this
    @type ardSupression: float
    @param forceSF: Force the prefactor in the kernel of the 
    GP to a particular value. Can give better results from the 
    GP interpolation.
    @type forceSF: float or None
    @param forceSigNoise: Parent in the experiment tree.
    @type forceSigNoise: ExperimentBranch,
    @param forceInterpolantRange: ARD sometimes ends up 
    ignoring time giving very odd results, here you can 
    force the GP to consider time a relevant independent 
    variable.
    @type forceInterpolantRange: float or None
    '''
    defaultTargets = ["*tMotionCaptureData","*tMarkerSet"]
    defaultIDExt = "com"
    def __init__(self, workingDirectory, analysisDirectory, metaData, parent):
        '''
        Constructor, loads suitable meta data after the parent class has had 
        a chance to set any defaults.
        @param workingDirectory: The working directory relative to which this objects 
        working directory should be resolved
        @type workingDirectory: string
        @param workingDirectory: The working directory relative to which this objects 
        working directory should be resolved
        @type workingDirectory: string
        @param metaData: Standard metadata dictionary.
        @type metaData: dict
        @param parent: Parent in the experiment tree.
        @type parent: ExperimentBranch,        
        '''
        BaseAnalysis.__init__(self, workingDirectory, analysisDirectory, metaData, parent)
        self.source = self.parent.resolve(self.targets[0])
        self.markers = self.parent.resolve(self.targets[1])
        self.sampleRate = self.source.sampleRate
        loadMetaData("numPC",self,self.metaData,3)
        loadMetaData("Nmeas",self,self.metaData,50)        
        loadMetaData("cached",self,self.metaData,True)
        loadMetaData("ardSupression",self,self.metaData,None)
        loadMetaData("forceSF",self,self.metaData,None)
        loadMetaData("forceSigNoise",self,self.metaData,None)
        loadMetaData("forceInterpolantRange",self,self.metaData,None)

    def _findCoMfromMeas(self):
        '''
        Compute the location of the center of mass from the body markers 
        and fill using gaussian process interpolation.
        If caching is turned on write the inferred CoM positions to disc.
        '''
        # Get marker positions on the body
        bodyMarkers = self.markers["BodyMarkers"]
        self.positions = np.vstack([self.source[b] for b in bodyMarkers])
        pos = np.array([self.positions[[j for j in range(self.positions.shape[0]) if j%3==i]] for i in range(3)])
        # Estimate the CoM where we can
        CoMEst = np.mean(pos,1)
        self.CoMUnfilled = np.array(CoMEst) # Diagnostic for checking interpolation
        # set up an array to store the 'times'
        tau = np.array(range(CoMEst.shape[1]))
        tau = tau*nanstd(CoMEst.reshape((CoMEst.size)))/nanstd(tau)
        # Find where we don't know the CoM position
        badIndex = np.isnan(CoMEst)
        if not np.sum(badIndex):
            self.CoM = CoMEst
            return; # If nothing needs to be filled, get out
        badIndexP = np.array([badIndex for i in range(pos.shape[1])]).transpose((1,0,2))
        goodIndex = ~np.isnan(CoMEst)  
        goodIndexP = np.array([goodIndex for i in range(pos.shape[1])]).transpose((1,0,2)) 
        # Detrend the CoM est so we get better estimates from the GP
        CoMTrend=list()
        for c,gI in zip(CoMEst,goodIndex):
            m, c, r, p, std_err = linregress(tau[gI],c[gI])
            CoMTrend.append(m*tau+c)
        CoMTrend = np.array(CoMTrend)
        CoMEst -= CoMTrend
        # Detrend the positions too
        npos = []
        for p in pos:
            pc = []
            for q in p:
                m, c, rv, pv, std_err = linregress(tau[gI],q[gI])
                pc.append(q - m*tau - c)
            npos.append(np.array(pc))
        pos = np.array(npos)
        tpos = pos[goodIndexP]
        bpos = pos[badIndexP]
        bpos = np.reshape(bpos,(pos.shape[0],pos.shape[1],bpos.shape[0]/(pos.shape[0]*pos.shape[1])))        
        pattern = np.isnan(bpos)
        pattern = pattern.transpose((0,2,1))
        upat = [[],[],[]]
        for i,p3 in enumerate(pattern):
            for p in p3:
                contains = False
                for up in upat[i]:
                    if sum((up-p)**2.0)==0:
                      contains = True
                if not contains:
                    upat[i].append(p)
        upat = np.array(upat).transpose((1,0,2))

        # Find regions where data is missing on a pattern of missing co-ordinate 
        # basis.
        for upt in upat:
            X = [p[goodIndex] for p in pos.transpose((1,0,2))]
            X = [p.reshape((pos.shape[0],p.shape[0]/pos.shape[0])) for p in X]
            X = np.array(X)
            X = X.transpose((1,0,2))[~upt]
            Xstar = [p[badIndex] for p in pos.transpose((1,0,2))]
            Xstar = [p.reshape((pos.shape[0],p.shape[0]/pos.shape[0])) for p in Xstar]
            Xstar = np.array(Xstar)
            Xstar = Xstar.transpose((1,0,2))[~upt]
            y = CoMEst[goodIndex]
            y = y.reshape(CoMEst.shape[0],y.shape[0]/CoMEst.shape[0])
            subGI = ~np.isnan(np.sum(Xstar,0))
            subXstar = Xstar.transpose()[subGI].transpose()
            for i,iy in enumerate(y):
                subIdx = np.array(range(CoMEst.shape[1]))
                subIdx = subIdx[badIndex[i]]
                subIdx = subIdx[subGI]
                oGI = np.array(range(CoMEst.shape[1]))
                oGI = oGI[goodIndex[i]]
                tau0 = tau[goodIndex[i]]
                tau1 = tau[badIndex[i]]
                tau1 = tau1[subGI]
                # Get a random sampling of points weighted to be near to the missing bits
                w = -(
                     np.repeat(oGI,subIdx.shape[0]).reshape((oGI.shape[0],subIdx.shape[0]))
                     - np.repeat(subIdx,oGI.shape[0]).reshape((subIdx.shape[0],oGI.shape[0])).transpose()
                    )**2
                w = np.sum(w,1)
                w -= np.max(w)
                w = w.astype(np.float32)
                w /= (self.Nmeas**2.0)*subIdx.shape[0] #nanstd(w.reshape(w.size))
                w = np.exp(w)
                w *= np.random.rand(w.size)
                w = sorted(zip(w,range(w.size)),key = lambda x: x[0])
                subsec = np.array(zip(*w[-self.Nmeas:])[1])
                # Do PCA to reduce the size of the input space
                if self.numPC>0 and X.shape[0]>=self.numPC:
                    '''
                replace prepca with...
from sklearn import decomposition
from sklearn import datasets
					pca = decomposition.PCA(n_components=3)
pca.fit(X)
X = pca.transform(X)'''
                    pca = decomposition.PCA()
                    pca.fit(X)
                    a = pca.transform(X)
                    #a = prepca(X)
                    fX = a[0][:self.numPC,:]
                    fsubXstar = np.dot(a[1],subXstar)[:self.numPC,:]
                    fsubXstar = np.vstack([tau1,fsubXstar])
                    fX = np.vstack([tau0,fX]) 
                elif self.numPC>0:
                    fX=X
                    fsubXstar = subXstar
                    fsubXstar = np.vstack([tau1,fsubXstar])
                    fX = np.vstack([tau0,fX]) 
                else:
                    fX = np.array([tau0])
                    fsubXstar = np.array([tau1])
                loghyper = np.append(10*np.ones(fX.shape[0]+1),0.1)
                # Interpolate with GP    
                gp = GPInterpolater(
                               fX.transpose()[subsec],
                               iy[subsec].reshape((iy[subsec].size,1)),
                               ['kernels.covSum', ['kernels.covSEard','kernels.covNoise']],
                               loghyper,
                               train=True
                              )
                if self.forceInterpolantRange:
                    gp.loghyper[0] = np.log(self.forceInterpolantRange)
                if self.ardSupression:
                    gp.loghyper[1:-3]= np.log(self.ardSupression*np.exp(gp.loghyper[1:-3]))
                if self.forceSF:
                    gp.loghyper[-2]= np.log(self.forceSF)
                if self.forceSigNoise:
                    gp.loghyper[-1]= np.log(self.forceSigNoise)
                CoMEst[i,subIdx] = gp.predict(fsubXstar.transpose())[0][:,0]
        CoMEst += CoMTrend
        self.CoM = CoMEst    
        if self.cached:
            # Write to disk
            pickle.dump(self.CoM, open(osjoin(self.analysisDirectory,"CoM.pickle"), 'wb'), -1)
            pickle.dump(self.CoMUnfilled, open(osjoin(self.analysisDirectory,"CoMUnfilled.pickle"), 'wb'), -1)

    def _findCoMfromDisc(self):
        '''
        Read the center of mass positions from disc.
        '''
        self.CoM = pickle.load(open(osjoin(self.analysisDirectory,"CoM.pickle"), 'r'))
        self.CoMUnfilled = pickle.load(open(osjoin(self.analysisDirectory,"CoMUnfilled.pickle"), 'r'))
        

    def _getCoM(self):
        '''
        If caching is turned off or no cache exists, fill in 
        the body markers using a Gaussian Process, then 
        compute the average to find an estimate of the CoM.
        Otherwise get the cached values
        '''
        if self.metaData["verbosity"]>0:
            print "Estimating the CoM, for CoM iD: %s Source iD: %s" % (self.iD,self.source.iD)
        if self.cached and "CoM.pickle" in listdir(self.analysisDirectory) and "CoMUnfilled.pickle" in listdir(self.analysisDirectory):
            self._findCoMfromDisc()
        else:
            self._findCoMfromMeas()

    def keys(self):
        '''
        The object computes CoM adjusted positions and the 
        CoM position itself and can be indexed to obtain 
        these values.
        '''
        return([k+"_CoMa" for k in self.source.keys()]+["CoM"])

    def has_key(self, k):
        '''
        Is the key k indexable? (Only true is k is either "CoM" or
        a CoM adjusted position. Just check is k is in self.keys()
        @param k: key
        @type k: any immutable
        '''
        return(k in self.keys())


    def __getitem__(self,k):
        '''
        Returns the value associated with key k. If 
        k is "CoM" return the CoM (calculating it if 
        necessary), otherwise if k is a CoM adjusted 
        position, calculate and return that. Otherwise 
        thow an exception.
        @param k: key
        @type k: any immutable
        '''
        # Mangle the key first
        if k=="CoM":
            if not self.__dict__.has_key("CoM"):
                self._getCoM()
            return self.CoM
        else:
            if k[-5:]=="_CoMa":
                if not self.__dict__.has_key("CoM"):
                    self._getCoM()
                return(self.source[k[:-5]]-self.CoM)
        raise KeyError("CoM analyses only contain information on the location of the Center of Mass! Invalid key %s",k)


    def _fillBodyMarkers(self):
        '''
        (!!! FIX THIS BIT !!!)
        Method takes the set of body markers and fills in regions where they are missing based on 
        where the other body markers are. Body markers are here markers expected to be somewhat rigid 
        in terms of their relative position. These are then used to construct an estimate for the 
        position of the center of mass in the method _getCenterOfMass.
        '''
        print "Filling Body Markers..."
        print "This may produce warnings where no markers were traced"
        # Get ourselves the labels for the filled markers and pair them 
        # the labels for the raw data.
        self.filledBodyLabels=["_".join(["filled",k]) for k in self.labels.bodyMarkers]
        markerPairs = [(p,q) for p in self.labels.bodyMarkers for q in self.labels.bodyMarkers]
        # Determine how each marker is offset from each other marker through time
        offsets  = np.array([nanmean(self[p]-self[q],1) for p,q in markerPairs])
        # Estimate a scale of the animal using the distance between each of these 
        # body markers.
        self.animalScaleEstimate=np.std(offsets)
        print "Approximate animal scale is ", self.animalScaleEstimate
        # For every position on the body compute a representative position by calculating where it is 
        # in relation to the other markers. Then when the marker position is unknown use the mean of this 
        # offset to fill in the missing marker position.
        offsets  = dict(zip(markerPairs,offsets))
        for p in self.labels.bodyMarkers:
            # Representative position calculated by using the offset position for every other marker
            representativePosition   = nanmean([(self[q].transpose()+offsets[(p,q)]).transpose() 
                                                for q in self.labels.bodyMarkers],0)
            # For every point in time we choose either a position from the representative positions or the 
            # measured position depending on if we know the measured position
            possiblePositions=zip(representativePosition.transpose(),self[p].transpose())
            self["_".join(["filled",p])] = TemporalSignal(
                                                               np.array([a if np.isnan(sum(b)) else b for a,b in possiblePositions]).transpose(),
                                                               self.time,
                                                               0)
        # All position labels should include the filled ones
        self.allPositionLabels.extend(self.filledBodyLabels)
  