'''
Created on 13 Feb 2011

@author: swilshin

Here we encapsulate the code for handling marker sets. The idea is there will be one generic 
class which will work under most circumstances, then you can plug in others when they are needed.
'''

from baseAnalysis import BaseAnalysis
from mDInterpret import getID

class LabelException(Exception):
    '''
    For when there is some screw up with labels that we really cant handle 
    such as there not being a standard form of the feet labels with which to 
    calculate phase.
    '''
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class MarkerSetInterface(BaseAnalysis):
    '''
    An interface which specifies what methods a marker set must have
    Note this is an inferace, not a class. It amounts to a promise to provide 
    certain methods and members (those pertaining to a weight), and it should 
    not be instantiated in raw form.
    '''
    defaultTargets = ["*tMotionCaptureData"]
    defaultIDExt = "ms"
    def __init__(self, workingDirectory, analysisDirectory, metaData, parent):
        BaseAnalysis.__init__(self, workingDirectory, analysisDirectory, metaData, parent)
        self.source = self.parent.resolve(self.targets[0])
        self.markerNames=self.source.keys()
        self._detectJunkMarkers()
        self._detectAppendages()
        self._detectBodyMarkers()
        self._detectHeadMarkers()
        self.appendageNames = []

    def getAppendages(self):
        '''
        Returns a list of legs for this marker set.  
        '''
        return(self.appendageMarkers.keys())
    
    def getAppendageMarkers(self, appendage):
        '''
        Returns a list of leg markers. These MUST be returned in the order the go along the leg, 
        with the first corresponding to either the toe or to the the marker furthest along the leg. 
        '''
        return(self.appendageMarkers[appendage])

    def _getAppendageTerminationMarker(self, appendage):
        '''
        Returns a the foot marker corresponding to a leg
        Made private to discourage use, since everything that uses this method ought to be done 
        by getting a hold of the leg in question
        '''
        return(self.getAppendageMarkers(appendage)[0])
    
    def _getAppendageTerminationMarkers(self):
        '''
        Returns a the foot marker corresponding to a leg
        Made private to discourage use, since everything that uses this method ought to be done 
        by getting a hold of the legs in question
        '''
        return([self.getAppendageMarkers(appendage)[0] for appendage in self.getAppendages()])

    def _detectJunkMarkers(self):
        '''
        Find junk markers, pure virtual method which always throws an exception if actually invoked.
        '''
        raise(conv.UnwrittenRoutine("Base virtual method in marker set interface: _detectJunkMarkers called, this method must be overridden!"))
        self.junkMarkers=[]
    
    def _detectAppendage(self):
        '''
        Find appendages, pure virtual method which always throws an exception if actually invoked.
        '''
        raise(conv.UnwrittenRoutine("Base virtual method in marker set interface: _detectAppendage called, this method must be overridden!"))
        self.appendageMarkers={}
    
    def _detectBodyMarkers(self):
        '''
        Find body markers, pure virtual method which always throws an exception if actually invoked.
        '''
        raise(conv.UnwrittenRoutine("Base virtual method in marker set interface: _detectBodyMarkers called, this method must be overridden!"))
        self.bodyMarkers=[]
    
    def _detectHeadMarkers(self):
        '''
        Find head markers, pure virtual method which always throws an exception if actually invoked.
        '''
        raise(conv.UnwrittenRoutine("Base virtual method in marker set interface: _detectHeadMarkers called, this method must be overridden!"))
        self.headMarkers=[]

    def _checkMarkerClassification(self):
        '''
        Check that all markers are classified, pure virtual method which always throws an exception if actually invoked.
        '''
        raise(conv.UnwrittenRoutine("Base virtual method in marker set interface: _checkMarkerClassification called, this method must be overridden!"))
    
    def soundOff(self):
        '''
        Report on identified markers, pure virtual method which always throws an exception if actually invoked.
        '''
        raise(conv.UnwrittenRoutine("Base virtual method in marker set interface: soundOff called, this method must be overridden!"))

    def keys(self):
        '''
        Return a list of different marker classifications that can be indexed.
        '''
        k = ["HeadMarkers","BodyMarkers","JunkMarkers"]
        k.extend([l+"Markers" for l in self.getAppendages()])
        k.append("Appendages")
        return(k)
        
    def has_key(self, k):
        '''
        Checks if k refers to an available set of markers or appendages, or is
        invalid.
        @param k: key
        @type k: any immutable
        '''
        return(k in self.keys())

    def __getitem__(self,k):
        '''
        If k refers to an available set of markers or appendages this will return it
        otherwise k is invalid and an exception is thrown.
        @param k: key
        @type k: any immutable
        '''
        if k == "HeadMarkers":
            return(self.headMarkers)
        if k == "BodyMarkers":
            return(self.bodyMarkers)
        if k == "JunkMarkers":
            return(self.bodyMarkers)
        for l in self.getAppendages():
            if l+"Markers" == k:
                return(self.appendageMarkers[l])
        if k == "Appendages":
            return(self.appendageNames)
        raise KeyError("Key %s not available in marker set."%k)
            
if __name__=="__main__":
    '''
    All classes defined here are abstract so only testing to be done is to check that 
    nothing foolish happens when the abstract classes are instantiated.
    '''