'''
Author: Simon Wilshin
Created: 11 Oct 2011
Cuts strides using phase
'''

'''
Author: Simon Wilshin
Created: 11 Oct 2011
'''

import numpy as np

from mDInterpret import getID
from baseAnalysis import BaseAnalysis
from mDInterpret import loadMetaData

class PhasicStrideCutter(BaseAnalysis):
    '''
    Takes a phase estimate and finds when it crosses 2*pi.
    Will find body and appendage phases. Default targets are a 
    local MarkerSet (to determine what legs are called) and a 
    local Phase.
    '''
    defaultTargets = ["*tMarkerSet","*tPhase"]
    defaultIDExt = "psc"
    def __init__(self,workingDirectory,analysisDirectory, metaData, parent):
        '''
        Constructor. Instantiate base class and resolve targets.
        @param workingDirectory: The working directory relative to which this objects 
        working directory should be resolved
        @type workingDirectory: string
        @param workingDirectory: The working directory relative to which this objects 
        working directory should be resolved
        @type workingDirectory: string
        @param metaData: Standard metadata dictionary.
        @type metaData: dict
        @param parent: Parent in the experiment tree.
        @type parent: ExperimentBranch,          
        '''
        BaseAnalysis.__init__(self, workingDirectory, analysisDirectory, metaData, parent)
        self.markers = self.parent.resolve(self.targets[0])
        self.phase = self.parent.resolve(self.targets[1])
        loadMetaData("cutPoint",self,self.metaData,np.pi/2.0)
        self.sampleRate = self.phase.sampleRate

    def has_key(self, k):
        '''
        Is k either a request for by strides or appendage strides. If so return true,
        otherwise return false. Just check if keys() contains k.
        @param k: Key for a set of strides
        @type k: string 
        '''
        return(k in self.keys())

    def keys(self):
        '''
        Give a list of keys that can be accessed, the Body strides and the Appendage strides.
        '''
        return(
               ["Body_Strides"]+["_".join([k,"Strides"]) for k in self.markers["Appendages"]]
              )

    def _Strides(self, k):
        '''
        Compute the strides for appendage k by finding when 
        the phase crosses 2*pi.
        @param k: Which appendages strides to compute
        @type k: string 
        '''
        ph = np.unwrap(self.phase[k+"_phase"][0])

        # Perform a moving average to smooth
        window=3
        shape = ph.shape[:-1] + (ph.shape[-1] - window + 1, window)
        strides = ph.strides + (ph.strides[-1],)
        sph = np.mean(np.lib.stride_tricks.as_strided(ph, shape=shape, strides=strides),1)%(2.0*np.pi)
        sph = (sph-self.cutPoint)%(2.0*np.pi)
        sph = np.insert(sph,0,sph[0])
        sph = np.append(sph,sph[-1])
        # Get the indicies of the strides
        c = sph[:-1]-sph[1:]>np.pi
        idx = np.array(range(c.size))[c]
        idx = np.insert(idx,0,0)
        idx = np.append(idx,c.size-1)
        self.data[k+"_Strides"] = zip(idx[:-1],idx[1:])
            
            
    def __getitem__(self,k):
        '''
        Returns an element of the dictionary self.data based on the key k 
        which is mangled (projected from a list of acceptable key names on 
        to the space of actual key names in data) first so the user has a 
        variety of sensible ways to get at the data. Will throw an exception 
        if k does not correspond to anything after mangling.
        Because certain quantities are computationally expensive to determine
        this method now checks to see if these have been calculated when 
        accessed. If they have not it calls the method to compute them. This 
        avoids the need to determine the at start up.
        Returns: The data associated with k.
        @param k: Which appendages strides to return
        @type k: string 
        '''
        # Mangle the key first
        nk=self.mangleKey(k)
        if (type(nk)==type(str()) or type(nk)==type(u"")):
            if not self.data.has_key(nk):
                if nk[-8:]=="_Strides":
                    self._Strides(nk[:-8])
        return(self.data[k])
