'''
Created on 05 Oct 2011

@author: Simon Wilshin, swilshin@rvc.ac.uk

Two classes, one which builds a phaser based on some data, another 
which links to that phasers and calculates phases for some data set.
'''

import numpy as np
import pylab as pl
from numpy import nanmean
from matplotlib import font_manager
from os.path import join as osjoin

from baseAnalysis import BaseAnalysis
from phaser import Phaser
from utils.protophases import VXKalmanPhase,VXKalmanPhasePSec
from mDInterpret import getID
from mDInterpret import loadMetaData
from utils import FiltFiltFilter

class PhSigProcessor(object):
    '''
    All the phaser based phase calculating classes need a way to arrange input signals 
    for phaser, this provides such an ability.
    '''
    def _makeSig(self,c,k,m):
       if k == "Body":
            '''
            t = [c[m[leg+"Markers"][0]+"_CoMa"][0] for leg in m["Appendages"]]
            t = [ti - nanmean(ti) for ti in t]
            t = t[0]-t[1]
            '''
            t=c["CoM"][2] # Use Z coordinate of CoM estimate.
            t -= nanmean(t)
            bf = FiltFiltFilter(c.source.sampleRate,3.0,3,"highpass")
            t = bf(t)
            t = np.reshape(t,(1,t.size))
       else:
            # Ask marker set for, e.g. "ForeLeftMarkers", then pull first (here "toeForeLeft") relative to CoMa -> then X coordinate. Once CoM GP stuff is done this will be quick.
            t = c[m[k+"Markers"][0]+"_CoMa"][0]
            t -= nanmean(t)
            t = np.reshape(t,(1,t.size))
       return t

class PhaserWrapper(BaseAnalysis,PhSigProcessor):
    '''
    Takes limb position estimates from some estimate of the 
    CoM and computes phases. Needs a marker set so it knows 
    what the limbs are. This is the phaser wrapper portion 
    of the phase estimate, it looks for local training data 
    (all targets from 1 forward) and trains a phaser using 
    this data for each limb. Use in conjunction with the 
    phase plug in below to evaluate the phase of individual 
    trials.
    Default targets are a local MarkerSet and a local CoM.
    '''
    defaultTargets = ["*tMarkerSet","*tCoM"]
    defaultIDExt = "phw"
    def __init__(self,workingDirectory,analysisDirectory, metaData, parent):
        '''
        Constructor, initialise a dictionary of phasers.
        @param workingDirectory: The working directory relative to which this objects 
        working directory should be resolved
        @type workingDirectory: string
        @param workingDirectory: The working directory relative to which this objects 
        working directory should be resolved
        @type workingDirectory: string
        @param metaData: Standard metadata dictionary.
        @type metaData: dict
        @param parent: Parent in the experiment tree.
        @type parent: ExperimentBranch,          
        '''
        BaseAnalysis.__init__(self, workingDirectory, analysisDirectory, metaData, parent)
        PhSigProcessor.__init__(self)
        self.markers = self.parent.resolve(self.targets[0])
        self.CoMs = [self.parent.resolve(t) for t in self.targets[1:]]
        self.sampleRate = self.CoMs[0].sampleRate
        loadMetaData("yscale",self,self.metaData,0.1)
        self.phasers = dict()

    def plotTrainingSignals(self,k):
        '''
        Plot the training signals for the quantity indexed k
        @param k: which limb to compute the phaser for.
        @type k: string   
        '''
        for c in self.CoMs:
            pl.figure()
            pl.title(c.iD)
            t = self._makeSig(c,k,self.markers)
            pl.plot(t.transpose())
            
       
    def _buildPhaser(self,k):
        '''
        Build our phaser for limb k.
        @param k: which limb to compute the phaser for.
        @type k: string          
        '''
        nt=[]
        self.phi = VXKalmanPhase(self.sampleRate, self.yscale)
        self.psi = VXKalmanPhasePSec(self.sampleRate, self.yscale)
        if self.metaData["verbosity"]>0:
            print "phase object iD: %s building phaser for limb %s" % (self.iD,k)
        for c in self.CoMs:
            if self.metaData["verbosity"]>0:
                print "  including CoM iD: %s" % c.iD
            t = self._makeSig(c,k,self.markers)
            nt.append(t)
        #self.phasers[k+"_Phaser"] = Phaser(nt, psecfunc = lambda x : np.dot([1],x),protophfun=self.phi)
        self.phasers[k+"_Phaser"] = Phaser(nt, psecfunc = self.psi, protophfun=self.phi) 

    def has_key(self,k):
        '''
        Check if one of the phasers we can construct is indexed by k.
        @param k: Index of desired phaser
        @type k: string 
        '''
        return(k in self.keys())

    def keys(self):
        return(["Body_Phaser"]+["_".join([k,"Phaser"]) for k in self.markers["Appendages"]])

    def __getitem__(self,k):
        '''
        Returns an element of the dictionary self.data based on the key k 
        which is mangled (projected from a list of acceptable key names on 
        to the space of actual key names in data) first so the user has a 
        variety of sensible ways to get at the data. Will throw an exception 
        if k does not correspond to anything after mangling.
        Because certain quantities are computationally expensive to determine
        this method now checks to see if these have been calculated when 
        accessed. If they have not it calls the method to compute them. This 
        avoids the need to determine the at start up.
        Returns: The data associated with k.
        '''
        # Mangle the key first
        nk=self.mangleKey(k)
        if (type(nk)==type(str()) or type(nk)==type(u"")):
            if nk[-7:]=="_Phaser":
                if not self.phasers.has_key(nk):
                    self._buildPhaser(nk[:-7])
        return(self.phasers[k])  
        
class Phase(BaseAnalysis,PhSigProcessor):
    '''
    Here we take a phase wrapper, link it to a marker set and a CoM 
    estimate and compute phase during the motion recorded in said 
    CoM estimate.
    '''
    defaultTargets = ["../*tPhaserWrapper","*tMarkerSet","*tCoM"]
    defaultIDExt = "phg"
    def __init__(self, workingDirectory, analysisDirectory, metaData, parent):
        BaseAnalysis.__init__(self, workingDirectory, analysisDirectory, metaData, parent)
        PhSigProcessor.__init__(self)
        self.phaser = self.parent.resolve(self.targets[0])
        self.markers = self.parent.resolve(self.targets[1])
        self.CoM = self.parent.resolve(self.targets[2])
        self.sampleRate = self.CoM.sampleRate
        
    def _getPhase(self,k):
        '''
        Calcualte the phase of object k.
        '''
        # Get the toe
        t = self._makeSig(self.CoM,k,self.markers)
        self.data[k+"_phase"] = self.phaser[k+"_Phaser"].phaserEval(t)%(2.0*np.pi)

    def has_key(self,k):
        return(k in self.keys())
        
    def keys(self):
        return(["Body_phase"] + ["_".join([k,"phase"]) for k in self.markers["Appendages"]])

    def __getitem__(self,k):
        '''
        Returns an element of the dictionary self.data based on the key k 
        which is mangled (projected from a list of acceptable key names on 
        to the space of actual key names in data) first so the user has a 
        variety of sensible ways to get at the data. Will throw an exception 
        if k does not correspond to anything after mangling.
        Because certain quantities are computationally expensive to determine
        this method now checks to see if these have been calculated when 
        accessed. If they have not it calls the method to compute them. This 
        avoids the need to determine the at start up.
        Returns: The data associated with k.
        '''
        # Mangle the key first
        nk=self.mangleKey(k)
        if (type(nk)==type(str()) or type(nk)==type(u"")):
            if nk[-6:]=="_phase":
                if not self.data.has_key(nk):
                    self._getPhase(nk[:-6])
        return(self.data[k])  

    def plotdiags(self,tofile=False,targetdir=None):
        if targetdir is None:
            targetdir=self.analysisDirectory
        # Phases:
        pl.figure()
        [pl.plot(self[k].T+2*np.pi*j,label=k) for j,k in enumerate(self.keys())]
        pl.legend(prop=font_manager.FontProperties(size=6))
        if tofile:
            # Self.iD, parent trial iD, Subject name
            print "  Writing image %s" % osjoin(targetdir, ".".join(["_".join([self.parent.parent.iD,self.parent.iD,self.iD,"phases"]),"pdf"]))
            pl.savefig(osjoin(targetdir, ".".join(["_".join([self.parent.parent.iD,self.parent.iD,self.iD,"phases"]),"pdf"])))
            pl.close()
        # Phase signals:
        pl.figure()
        [pl.plot(self._makeSig(self.CoM,k[:-6],self.markers).T+2*np.pi*j,label=k[:-6]) for j,k in enumerate(self.keys())]
        pl.legend(prop=font_manager.FontProperties(size=6))
        if tofile:
            # Self.iD, parent trial iD, Subject name
            print "  Writing image %s" % osjoin(targetdir, ".".join(["_".join([self.parent.parent.iD,self.parent.iD,self.iD,"phasesigs"]),"pdf"]))
            pl.savefig(osjoin(targetdir, ".".join(["_".join([self.parent.parent.iD,self.parent.iD,self.iD,"phasesigs"]),"pdf"])))
            pl.close()
        

        