'''
Created on 04 Oct 2011

@author: Simon Wilshin

The data in the sml stride analyse library is stored in a bidirectional 
tree, this implements the tree class
'''

class TreeException(Exception):
    '''
    For when you try to access outside of the tree, or add something 
    invalid to the tree.
    Records if the exception resulted in a resolvable end point or not. 
    Adding an invalid addition to the tree is considered resolvable, since 
    in principle there is something one can access at the id requested.
    Attempting to make a leaf resolve also is considered irresolvable 
    as one has dropped off the tree.
    '''
    def __init__(self, value, resolvable=True):
        self.value = value
        self.resolvable = resolvable
    def __str__(self):
        return repr(self.value)

class ID(object):
    '''
    A class for storing unique IDs. Will auto assign 
    if none provided using a static member
    '''
    usedIDs = []
    currentID = 0
    def __init__(self, iD=None):
        if iD==None:
            while(any([i==str(ID.currentID) for i in ID.usedIDs])):
                ID.currentID+=1
            iD=ID.currentID
        if any([i==str(iD) for i in ID.usedIDs]):
            raise TreeException("ID %s already in use"%iD)
        else:
            self.iD = str(iD)
            ID.usedIDs.append(str(iD))

class Branch(ID):
    '''
    A branch of a bidirectional tree.
    '''
    def __init__(self,parent=None, iD=None):
        ID.__init__(self,iD)
        self.parent = parent
        self.children = []
        if parent!=None:
            self.parent.addChild(self)

    def addChild(self, child):
        for c in self.children:
            if c.iD == child.iD:
                raise TreeException("All entries at a branch of the tree must have unique ID. Child id %s"%str(child.iD),False)
        self.children.append(child)

    def canResolve(self,dest):
        resolvable = True
        try:
            self.resolve(dest)
        except TreeException as e:
            resolvable = e.resolvable
        return(resolvable)

    def resolve(self,dest):
        if type(dest)!=list:
            dest = dest.split("/")
        if len(dest)==1:
            if dest[0][:2] == "*t":
                possDests = [c for c in self.children if c.__class__.__name__ == dest[0][2:]]
                if len(possDests) == 1:
                    return(possDests[0])
                elif len(possDests) == 0:
                    raise TreeException("No matching entry.",False)
                else:
                    raise TreeException("Resolving in tree ambiguous, multiple possible destinations for " + str(dest) + " in " + str(self.iD) + ".")
            else:
                for c in self.children:
                    if c.iD == dest[0]:
                        return(c)
        elif dest[0]=="..":
            if self.parent==None:
                raise TreeException("No parent beyond root.",False)
            return(self.parent.resolve(dest[1:]))
        else:
            if dest[0][:2] == "*t":
                possDests = [c for c in self.children if c.__class__.__name__ == dest[0][2:]]
                if len(possDests) == 1:
                    return(possDests[0].resolve(dest[1:]))
                elif len(possDests) == 0:
                    raise TreeException("No matching entry.",False)
                else:
                    raise TreeException("Resolving in tree ambiguous, multiple possible destinations.")
            else:
                for c in self.children:
                    if c.iD == dest[0]:
                        return(c.resolve(dest[1:]))            
            raise TreeException("Invalid tree syntax, could not resolve %s"%str(dest))

class Leaf(ID):
    '''
    A leaf in the tree
    '''
    def __init__(self,parent=None, iD=None):
        ID.__init__(self,iD)
        self.parent=parent
        self.children = None
        if parent!=None:
            self.parent.addChild(self)

    def resolve(self,dest):
        raise TreeException("Attempt to make a leaf resolve!", False)

if __name__=="__main__":
    '''
    Any simple tests of this framework go here.
    '''
    ID(3)
    ID("Fish")
    ID()
    ID()
    ID()
    ID()
    r = Branch(iD = "Root")
    b1 = Branch(r,"B1")
    b2 = Branch(b1, "B2")
    b3 = Branch(r,"B3")
    l = Leaf(b3,"L")
    print "Test 1: ", ID.usedIDs == ['3', 'Fish', '0', '1', '2', '4', 'Root', 'B1', 'B2', 'B3', 'L']
    print "Test 2: ", ID.currentID == 4    
    print "Test 3: ", b2.resolve("../../B3/L").iD == "L"
    print "Test 4: ", b2.resolve("../../B3/*tLeaf").iD == "L"
    print "Test 5: ", b3.resolve("../B1").iD == "B1"
    print "Test 6: ", b3.resolve("../B1/*tBranch").iD == "B2"
    print "Test 7: ", b2.canResolve("../../B3/L") == True
    print "Test 8: ", b2.canResolve("../../../B3/L") == False
    print "Test 9: ", b2.canResolve("../../B3/L/F") == False
    
