'''
Created on 12 Nov 2010

@author: Simon Wilshin

Various unsorted utility functions including a method for finding flat regions in an array, 
a method for removing nans from an array and a method for sanitising strings which contain 
whitespace.

Another function interpolates values that are recorded as NaNs.

Another pair of functions for quadratic peak finding

Another function for simple peak finding in nice smooth data.

'''

import os
import numpy as np
from matplotlib.mlab import find
import scipy.stats

def ensure_dir(dir):
    if not os.path.isdir(dir):
        os.makedirs(dir)

def smlDictToSmlMat(dic):
    '''
    Takes a dictionary organised as a list of dictionaries of lists and makes it 
    into a dictionary of matricies, used primarily to arrange lists of signals so 
    they can be averaged.
    '''
    absdByt=dict()
    for dat in dic:
        if absdByt==dict():
            for d in dat:
                absdByt[d]=list()
        for d in dat:
            absdByt[d].append(dat[d])
    return(dict([(d,np.array(absdByt[d])) for d in absdByt]))

def nancounts(data):
    '''
    Counts the number of nans in the columns of data.
    Input: NxM numpy array.
    Output: 1xM numpy array, with counts of NaNs in each column.
    '''
    nval=nans(data.shape[1])
    for k in range(data.shape[1]):
        nval[k]=sum(elemnot(np.isnan(data[:,k])))
    return nval

def nanpctile(data,pct):
    '''
    Returns percentiles of the given data, columnwise, ignoring NaN elements.
    WARNING: May not be well-behaved for small numbers of elements!!!!
    Expects a NxM numpy array and returns a 1xM array, computing pctile down
    the first dimension (0).
    '''
    sc=nans(data.shape[1])
    for k in range(data.shape[1]):
        sc[k]=scipy.stats.scoreatpercentile(data[elemnot(np.isnan(data[:,k])),k],pct)
    return sc

def norm(x):
    return(np.sqrt(np.sum([np.array(x[k])**2 for k in range(0,3)],0)))

def eucnorm( data, axis=1 ):
    '''Computes the Euclidean norm of the given numpy array. operates along
    dimension 1 by default. This is columns in a 2D array.'''
    return ((np.sum(data**2.0,axis))**0.5)

def nans(size):
    '''Creates a numpy array of the given size filled with NaNs'''
    foo = np.empty(size)
    foo.fill(np.nan)
    return(foo)

def elemnot(data):
    '''Computes the elementwise logical NOT of the given item''' 
    return([not k for k in data])

def fillNans(datain):
    '''Uses linear interpolation to fill in Nans in the given array.
    If given >1 dimensional data, fills columns (down first dimension)
    in turn.'''
    data = np.array(datain)
    if data.ndim > 1:
        for k in range(data.shape[1]):
            bad = find(np.isnan(data[0:,k]))
            good = find([not k for k in np.isnan(data[0:,k])])
            data[bad,k] = np.interp(bad,good,data[good,k]) 
    else:
        bad = find(np.isnan(data))
        good = find([not k for k in np.isnan(data)])
        data[bad] = np.interp(bad,good,data[good]) 
    return(data)

# Testing
eps=np.finfo(type(1.0)).eps

def dot(x,y):
    return(np.sum([np.array(x[k])*np.array(y[k]) for k in range(0,3)],0))

def angle(x,y):
    t_dot=dot(x,y)/(norm(x)*norm(y))
    if np.array(t_dot).shape==():
        if t_dot>1.0:
            t_dot=1.0
        if t_dot<-1.0:
            t_dot=-1.0
    else:
        t_dot=np.array([tau if np.abs(tau)<=1 else np.sign(tau) for tau in t_dot])
    return(np.arccos(t_dot))

def testDot():
    print "Testing the dot function."
    t1 = dot([1.0,1.0,1.0],[1.0,1.0,1.0])<3.0+2.0*eps and dot([1.0,1.0,1.0],[1.0,1.0,1.0])>3.0-2.0*eps
    print "Test 0001 Passed: ", t1
    t2 = dot([[1.0,1.0],[1.0,1.0],[1.0,1.0]],[[1.0,1.0],[1.0,1.0],[1.0,1.0]])[1]<3.0+2.0*eps and dot([[1.0,1.0],[1.0,1.0],[1.0,1.0]],[[1.0,1.0],[1.0,1.0],[1.0,1.0]])[1]>3.0-2.0*eps
    print "Test 0002 Passed: ", t2
    testsPassed=t1 and t2
    return(testsPassed)

def testAngle():
    print "Testing the angle function."
    t1 = angle([1.0,1.0,1.0],[1.0,1.0,1.0])<0.0+2.0*eps and angle([1.0,1.0,1.0],[1.0,1.0,1.0])>0.0-2.0*eps
    print "Test 0001 Passed: ", t1
    t2 = angle([[1.0,1.0],[1.0,1.0],[1.0,1.0]],[[1.0,1.0],[1.0,1.0],[1.0,1.0]])[1]<0.0+2.0*eps and angle([[1.0,1.0],[1.0,1.0],[1.0,1.0]],[[1.0,1.0],[1.0,1.0],[1.0,1.0]])[1]>0.0-2.0*eps
    print "Test 0002 Passed: ", t2
    t3 = angle([1.0,0.0,0.0],[0.0,0.0,1.0])<np.pi/2.0+2.0*eps and angle([1.0,0.0,0.0],[0.0,0.0,1.0])>np.pi/2.0-2.0*eps
    print "Test 0003 Passed: ", t3   
    testsPassed=t1 and t2 and t3
    return(testsPassed)   

def testNorm():
    print "Testing the norm function."
    t1 = norm([2.0,2.0,2.0])<np.sqrt(12.0)+2.0*eps and norm([2.0,2.0,2.0])>np.sqrt(12.0)-2.0*eps
    print "Test 0001 Passed: ", t1
    t2 = norm([0.0,0.0,0.0])<2.0*eps and norm([0.0,0.0,0.0])>-2.0*eps
    print "Test 0002 Passed: ", t2
    t3 = norm(np.array([[0.0,1.0],[0.0,1.0],[0.0,1.0]]))[0]<2.0*eps and norm(np.array([[0.0,1.0],[0.0,1.0],[0.0,1.0]]))[0]>-2.0*eps
    print "Test 0003 Passed: ", t3
    t4 = norm(np.array([[0.0,1.0],[0.0,1.0],[0.0,1.0]]))[1]<np.sqrt(3)+2.0*eps and norm(np.array([[0.0,1.0],[0.0,1.0],[0.0,1.0]]))[1]>np.sqrt(3)-2.0*eps
    print "Test 0004 Passed: ", t4
    testsPassed=t1 and t2 and t3 and t4
    return(testsPassed)

def testUtils():
    print "Testing utils module, True indicates test passed."      
    t1=testNorm()
    t2=testDot()
    t3=testAngle()
    testsPassed=t1 and t2 and t3
    if testsPassed:
        print "All tests PASSED in module utils."
    else:
        print "Some tests FAILED in module utils!"
    return(testsPassed)
    
if __name__=='__main__':
    '''
    Run some tests, should print True if everything worked
    '''
    testUtils()